class QgsApplication:
    @staticmethod
    def instance():
        return None


class QgsMesh:
    pass


class QgsMeshLayer:
    pass


class QgsMeshSpatialIndex:
    pass


class QgsPointXY:
    pass


class QgsPoint:
    pass


class QgsGeometry:
    pass


class QgsMeshDatasetIndex:
    pass


class QgsProject:
    pass


class QgsDistanceArea:
    pass


class QgsMeshDatasetGroupMetadata:
    pass
